#ifndef lint
static char sccsid[] = "@(#)bootp.c	1.1 (Stanford) 1/22/86";
static char rcsid[] = "$Header: bootpd.c,v 1.31 93/08/29 20:46:45 leres Exp $";
#endif


/*
 * BOOTP (bootstrap protocol) server daemon.
 *
 * Answers BOOTP request packets from booting client machines.
 * See [SRI-NIC]<RFC>RFC951.TXT for a description of the protocol.
 * See [SRI-NIC]<RFC>RFC1048.TXT for vendor-information extensions.
 * See accompanying man page -- bootpd.8
 *
 *
 * Copyright 1988 by Carnegie Mellon.
 *
 * Permission to use, copy, modify, and distribute this program for any
 * purpose and without fee is hereby granted, provided that this copyright
 * and permission notice appear on all copies and supporting documentation,
 * the name of Carnegie Mellon not be used in advertising or publicity
 * pertaining to distribution of the program without specific prior
 * permission, and notice be given in supporting documentation that copying
 * and distribution is by permission of Carnegie Mellon and Stanford
 * University.  Carnegie Mellon makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 *
 * HISTORY
 *
 * 01/22/86	Bill Croft at Stanford University
 *		    Created.
 *
 * 07/30/86     David Kovar at Carnegie Mellon University
 *		    Modified to work at CMU.
 *
 * 07/24/87	Drew D. Perkins at Carnegie Mellon University
 *		    Modified to use syslog instead of Kovar's
 *		    routines.  Add debugging dumps.  Many other fixups.
 *
 * 07/15/88	Walter L. Wimer at Carnegie Mellon University
 *		    Added vendor information to conform to RFC1048.
 *		    Adopted termcap-like file format to support above.
 *		    Added hash table lookup instead of linear search.
 *		    Other cleanups.
 *
 *
 * BUGS
 *
 * Currently mallocs memory in a very haphazard manner.  As such, most of
 * the program ends up core-resident all the time just to follow all the
 * stupid pointers around. . . .
 *
 */




#include <sys/types.h>
#ifndef __P
#include <sys/cdefs.h>
#endif
#include <sys/file.h>
#include <sys/ioctl.h>
#ifdef SUNOS40
#include <sys/sockio.h>
#endif
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <net/if.h>
#ifdef SUNOS40
#include <net/if_arp.h>
#endif

#include <netinet/in.h>

#include <arpa/inet.h>

#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <strings.h>
#include <syslog.h>
#include <unistd.h>

#include "bootp.h"
#include "hash.h"
#include "bootpd.h"

#define HASHTABLESIZE 257	/* Hash table size (prime) */
#define DEFAULT_TIMEOUT 15L	/* Default timeout in minutes */
#define MAXHOPS 5		/* Ignore packets after this many hops */

#ifndef CONFIG_FILE
#define CONFIG_FILE		"/etc/bootptab"
#endif
#ifndef DUMP_FILE
#define DUMP_FILE		"/etc/bootpd.dump"
#endif

#define MAX(a, b) ((a) < (b)? (b): (a))



/*
 * Externals, forward declarations, and global variables
 */

extern char Version[];
extern int errno;
extern char *malloc __P((u_int));

int	chk_access __P((char *, long *));
void	dovend_cmu __P((struct bootp *, struct host *));
void	dovend_rfc1048 __P((struct bootp *, struct host *, long));
#ifdef DEBUG
void	dump_host __P((FILE *f, struct host *));
void	dumptab __P((void));
#endif
void	forwardrequest __P((struct host *));
boolean	hwlookcmp __P((struct host *, struct host *));
void	insert_generic __P((struct shared_bindata *, byte **, int *));
void	insert_ip __P((int, struct addr_list *, byte **, int *));
void	insert_u_long __P((u_long, byte **));
#ifdef DEBUG
void	list_ipaddresses __P((FILE *, struct addr_list *));
#endif
int	nmatch __P((u_long, u_long));
void	process __P((struct bootp *, int, int));
void	reply __P((void));
void	request __P((void));
void	sendreply __P((struct host *));
int	setarp __P((u_long, byte *, int));
void	usage __P((void));

/*
 * IP port numbers for client and server obtained from /etc/services
 */

u_short bootps_port, bootpc_port;


/*
 * Internet socket and interface config structures
 */

static struct sockaddr_in sin;
struct sockaddr_in from;	/* Packet source */
struct ifreq ifreq[10];		/* Holds interface configuration */
struct ifconf ifconf;		/* Int. config ioctl block (pnts to ifreq) */
struct arpreq arpreq;		/* Arp request ioctl block */


/*
 * General
 */

int debug = 0;			/* Debugging flag */
int quiet = 0;			/* quiet mode */
int server_s;			/* server port socket file descriptor */
int client_s;			/* client port socket file descriptor */
byte *buf;			/* Receive packet buffer */
u_int bufsize = 1024;		/* malloc'ed size of buf */
struct timezone tzp;		/* Time zone offset for clients */
struct timeval tp;		/* Time (extra baggage) */
long secondswest;		/* Time zone offset in seconds */

/*
 * Globals below are associated with the bootp database file (bootptab).
 */

char *bootptab = NULL;
#ifdef DEBUG
char *bootpd_dump = NULL;
#endif



/*
 * Vendor magic cookies for CMU and RFC1048
 */

u_char vm_cmu[4] = VM_CMU;
u_char vm_rfc1048[4] = VM_RFC1048;


/*
 * Hardware address lengths (in bytes) based on hardware type code.
 * List in order specified by Assigned Numbers RFC; Array index is
 * hardware type code.  Entries marked as zero are unknown to the author
 * at this time. . . .
 */

u_int maphaddrlen[MAXHTYPES + 1] = {
     0,		/* Type 0: Reserved (don't use this) */
     6,		/* Type 1: 10Mb Ethernet (48 bits) */
     1,		/* Type 2: 3Mb Ethernet (8 bits) */
     0,		/* Type 3: Amateur Radio AX.25 */
     1,		/* Type 4: Proteon ProNET Token Ring */
     0,		/* Type 5: Chaos */
     6,		/* Type 6: IEEE 802 Networks */
     0		/* Type 7: ARCNET */
};


/*
 * Main hash tables
 */

hash_tbl *hwhashtable;
hash_tbl *iphashtable;
hash_tbl *nmhashtable;




/*
 * Initialization such as command-line processing is done and then the main
 * server loop is started.
 */

int
main(argc, argv)
	int argc;
	char **argv;
{
	struct timeval actualtimeout, *timeout;
	struct bootp *bp;
	struct servent *servp;
	char *stmp;
	int n, tolen, fromlen;
	int nfound, readfds;
	int standalone;

	stmp = NULL;
	standalone = FALSE;
	actualtimeout.tv_usec = 0L;
	actualtimeout.tv_sec  = 60 * DEFAULT_TIMEOUT;
	timeout = &actualtimeout;

	/*
	 * Read switches.
	 */
	for (argc--, argv++; argc > 0; argc--, argv++) {
		if (argv[0][0] == '-') {
			switch (argv[0][1]) {

			case 't':
				if (argv[0][2]) {
					stmp = &(argv[0][2]);
				} else {
					argc--;
					argv++;
					stmp = argv[0];
				}
				if (!stmp || (sscanf(stmp, "%d", &n) != 1) || (n < 0)) {
				    fprintf(stderr,
					    "bootpd: invalid timeout specification\n");
				    break;
				}
				actualtimeout.tv_sec = (long) (60 * n);
				/*
				 * If the actual timeout is zero, pass a NULL pointer
				 * to select so it blocks indefinitely, otherwise,
				 * point to the actual timeout value.
				 */
				timeout = (n > 0) ? &actualtimeout : NULL;
				break;

			case 'd':
				debug++;
				break;

			case 's':
				standalone = TRUE;
				break;

			case 'q':
				quiet++;
				break;

			default:
				fprintf(stderr,
					"bootpd: unknown switch: -%c\n",
					argv[0][1]);
				usage();
				break;
			}
		} else {
		if (!bootptab) {
		    bootptab = argv[0];
#ifdef DEBUG
		} else if (!bootpd_dump) {
		    bootpd_dump = argv[0];
#endif
		} else {
		    fprintf(stderr, "bootpd: unknown argument: %s\n", argv[0]);
		    usage();
		}
	    }
	}

	/* Malloc receive buffer */
	buf = malloc(bufsize);
	if (buf == NULL) {
		syslog(LOG_ERR, "receive buffer malloc failed: %m");
		exit(1);
	}
	bp = (struct bootp *) buf;

	/*
	 * Set default file names if not specified on command line
	 */
	if (!bootptab)
		bootptab = CONFIG_FILE;
#ifdef DEBUG
	if (!bootpd_dump)
		bootpd_dump = DUMP_FILE;
#endif

	/*
	 * Get our timezone offset so we can give it to clients if the
	 * configuration file doesn't specify one.
	 */
	if (gettimeofday(&tp, &tzp) < 0) {
		/* Assume GMT for lack of anything better */
		secondswest = 0L;
		syslog(LOG_ERR, "gettimeofday: %m");
	} else {
		/* Convert to seconds */
		secondswest = 60L * tzp.tz_minuteswest;
	}

	/*
	 * Allocate hash tables for hardware address, ip address, and hostname
	 */
	hwhashtable = hash_Init(HASHTABLESIZE);
	iphashtable = hash_Init(HASHTABLESIZE);
	nmhashtable = hash_Init(HASHTABLESIZE);
	if (!(hwhashtable && iphashtable && nmhashtable)) {
		syslog(LOG_ERR, "Unable to allocate hash tables");
		exit(1);
	}

	if (standalone) {
		/*
		 * Go into background and disassociate from
		 * controlling terminal.
		 */
		if (debug < 3) {
			if (fork())
				exit(0);
			(void) setpgrp(0, getpid());
			(void) chdir("/");
			for (n = 0; n < 10; n++)
				(void) close(n);
			(void) open("/", O_RDONLY);
			(void) dup2(0, 1);
			(void) dup2(0, 2);
			n = open("/dev/tty", O_RDWR);
			if (n >= 0) {
				ioctl(n, TIOCNOTTY, (char *) 0);
				(void) close(n);
			}
		}
		/*
		 * Nuke any timeout value
		 */
		timeout = NULL;
	}


	/*
	 * Initialize logging.
	 */
	openlog("bootpd", 0, LOG_DAEMON);
#ifdef notdef
	syslog(LOG_INFO, "%s", Version);
#endif

	/* Get a client socket */
	if ((client_s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		syslog(LOG_ERR, "fwd socket: %m");
		exit(1);
	}

	/* Obtain bootp client port numbers  */
	servp = getservbyname("bootpc", "udp");
	if (servp) {
		bootpc_port = ntohs((u_short) servp->s_port);
	} else {
		bootpc_port = (u_short) IPPORT_BOOTPC;
		syslog(LOG_ERR,
		    "udp/bootpc: unknown service -- assuming port %d",
		    bootpc_port);
	}

	/* Bind client socket */
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_port = htons(bootpc_port);
	if (bind(client_s, &sin, sizeof(sin)) < 0) {
		syslog(LOG_ERR, "client port bind: %m");
		exit(1);
	}

	if (standalone) {
		/* Get a server socket */
		if ((server_s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
			syslog(LOG_ERR, "socket: %m");
			exit(1);
		}

		/* Obtain bootp server port number.  */
		servp = getservbyname("bootps", "udp");
		if (servp) {
			bootps_port = ntohs((u_short) servp->s_port);
		} else {
			bootps_port = (u_short) IPPORT_BOOTPS;
			syslog(LOG_ERR,
			    "udp/bootps: unknown service -- assuming port %d",
			    bootps_port);
		}

		/* Bind server socket */
		sin.sin_family = AF_INET;
		sin.sin_addr.s_addr = INADDR_ANY;
		sin.sin_port = htons(bootps_port);
		if (bind(server_s, &sin, sizeof(sin)) < 0) {
			syslog(LOG_ERR, "server port bind: %m");
			exit(1);
		}
	} else {
		/*
		 * Assume socket was passed to us from inetd
		 */
		server_s = 0;
		tolen = sizeof(sin);
		bzero((char *) &sin, tolen);
		if (getsockname(server_s, &sin, &tolen) < 0) {
			syslog(LOG_ERR, "getsockname: %m");
			exit(1);
		}
		/* Use bootp server port inetd passed us */
		bootps_port = ntohs(sin.sin_port);
	}


	/*
	 * Determine network configuration.
	 */
	ifconf.ifc_len = sizeof(ifreq);
	ifconf.ifc_req = ifreq;
	if ((ioctl(server_s, SIOCGIFCONF, (caddr_t) &ifconf) < 0) ||
	    (ifconf.ifc_len <= 0)) {
		syslog(LOG_ERR, "ioctl: %m");
		exit(1);
	}

	/*
	 * Read the bootptab file once immediately upon startup.
	 */
	readtab();

	/*
	 * Set up signals to read or dump the table.
	 */
	if ((int) signal(SIGHUP, readtab) < 0) {
		syslog(LOG_ERR, "signal: %m");
		exit(1);
	}
#ifdef DEBUG
	if ((int) signal(SIGUSR1, dumptab) < 0) {
		syslog(LOG_ERR, "signal: %m");
		exit(1);
	}
#endif

	/*
	 * Process incoming requests.
	 */
	for (;;) {
		readfds = (1 << client_s) + (1 << server_s);
		nfound = select(MAX(client_s, server_s) + 1,
		    &readfds, NULL, NULL, timeout);
		if (nfound < 0) {
			if (errno != EINTR) {
				syslog(LOG_ERR, "select: %m");
			}
			continue;
		}

		/* Receive server packet */
		if (readfds & (1 << server_s)) {
			fromlen = sizeof(from);
			n = recvfrom(server_s, buf, bufsize, 0,
			    &from, &fromlen);
			process(bp, n, 1);	/* isserver true */
			continue;
		}

		/* Receive client packet */
		if (readfds & (1 << client_s)) {
			fromlen = sizeof(from);
			n = recvfrom(client_s, buf, bufsize, 0,
			    &from, &fromlen);
			process(bp, n, 0);	/* isserver false */
			continue;
		}

		/* If we get here, we timed out */
#ifdef notdef
		syslog(LOG_INFO, "exiting after %ld minutes of inactivity",
		    actualtimeout.tv_sec / 60);
#endif
		exit(0);
	}
}

void
process(bp, n, isserver)
	register struct bootp *bp;
	int n;
	int isserver;
{
	if (n <= 0) {
		if (debug)
			syslog(LOG_ERR, "received empty packet");
		return;
	}
	if (n < sizeof(struct bootp)) {
		if (debug)
			syslog(LOG_ERR, "received short packet");
		return;
	}

	/* Maybe re-read bootptab */
	readtab();

	switch (bp->bp_op) {

	case BOOTREQUEST:
		if (!isserver) {
			/* as per the spec */
			syslog(LOG_ERR, "discarding request from client port");
			break;
		}
		request();
		break;

	case BOOTREPLY:
		reply();
		break;

	default:
		if (debug)
			syslog(LOG_ERR, "discarding unknown %s op %d",
			    isserver ?  "server" : "client", bp->bp_op);
	}
}




/*
 * Print "usage" message and exit
 */

void
usage()
{
    fprintf(stderr,
	"usage:  bootpd [-d] [-s] [-q] [-t timeout] [configfile [dumpfile]]\n");
    fprintf(stderr, "\t -d\tincrease debug verbosity\n");
    fprintf(stderr, "\t -s\trun standalone (without inetd)\n");
    fprintf(stderr, "\t -q\tquiet mode\n");
    fprintf(stderr, "\t -t n\tset inetd exit timeout to n minutes\n");
    exit(1);
}



/*
 * Process BOOTREQUEST packet.
 *
 * (This version does not interpret the hostname field of the
 * request packet;  it COULD do a name->address lookup and
 * forward the request there.)
 */
void
request()
{
    register struct bootp *bp;
    register struct host *hp;
    register int n;
    char path[128];
    struct host dummyhost;
    long bootsize;
    u_int hlen, hashcode;

    bp = (struct bootp *) buf;
    if (bp->bp_ciaddr == 0) {
	/*
	 * client doesnt know his IP address,
	 * search by hardware address.
	 */
	if (debug)
	    syslog(LOG_DEBUG, "request from hardware address %s",
		haddrtoa(bp->bp_chaddr, bp->bp_htype));

	dummyhost.htype = bp->bp_htype;
	hlen = haddrlength(bp->bp_htype);
	bcopy(bp->bp_chaddr, dummyhost.haddr, hlen);
	hashcode = hash_HashFunction(bp->bp_chaddr, hlen);
	hp = (struct host *) hash_Lookup(hwhashtable, hashcode, hwlookcmp,
					 &dummyhost);
	if (hp == NULL) {
	    if (quiet == 0)
		syslog(LOG_NOTICE, "hardware address not found: %s",
		    haddrtoa(bp->bp_chaddr, bp->bp_htype));
	    return;
	}
	bp->bp_yiaddr = hp->iaddr;

    } else {

	/*
	 * search by IP address.
	 */
	if (debug)
	    syslog(LOG_DEBUG, "request from IP addr %s", intoa(bp->bp_ciaddr));
	dummyhost.iaddr = bp->bp_ciaddr;
	hashcode = hash_HashFunction((u_char *) &(bp->bp_ciaddr), 4);
	hp = (struct host *) hash_Lookup(iphashtable, hashcode, iplookcmp,
					 &dummyhost);
	if (hp == NULL) {
	    if (quiet == 0)
		syslog(LOG_NOTICE, "IP address not found: %s",
		    intoa(bp->bp_ciaddr));
	    return;
	}
    }

    if (debug) {
	if (hp->flags.iaddr)
	    syslog(LOG_INFO, "found %s (%s)",
		hp->hostname->string, intoa(hp->iaddr));
	else
	    syslog(LOG_INFO, "found %s", hp->hostname->string);
    }

    /* Discard packet if not enough seconds have passed */
    if (hp->flags.ignore_secs && bp->bp_secs <= hp->ignore_secs)
	    return;

    /*
     * Check to see if we want to forward this request.
     */
    if (hp->flags.forward) {
	    forwardrequest(hp);
	    return;
    }

    /*
     * Fill in the client's proper bootfile.
     *
     * If the client specifies an absolute path, try that file with a
     * ".host" suffix and then without.  If the file cannot be found, no
     * reply is made at all.
     *
     * If the client specifies a null or relative file, use the following
     * table to determine the appropriate action:
     *
     *  Homedir      Bootfile    Client's file
     * specified?   specified?   specification   Action
     * -------------------------------------------------------------------
     *      No          No          Null         Send null filename
     *      No          No          Relative     Discard request
     *      No          Yes         Null         Send if absolute else null
     *      No          Yes         Relative     Discard request
     *      Yes         No          Null         Send null filename
     *      Yes         No          Relative     Lookup with ".host"
     *      Yes         Yes         Null         Send home/boot or bootfile
     *      Yes         Yes         Relative     Lookup with ".host"
     *
     */

    /*
     * Who knows. . . . . vestigial code (from Stanford???) which should
     * probably be axed.
     */
#ifdef notdef
    if (strcmp(bp->bp_file, "sunboot14") == 0)
	bp->bp_file[0] = 0;	/* pretend it's null */
#endif

    /*
     * This is the code which actually implements the above truth table.
     */
    if (bp->bp_file[0]) {
	/*
	 * The client specified a file.
	 */
	if (bp->bp_file[0] == '/') {
	    strcpy(path, bp->bp_file);		/* Absolute pathname */
	} else {
	    if (hp->flags.homedir) {
		strcpy(path, hp->homedir->string);
		strcat(path, "/");
		strcat(path, bp->bp_file);
	    } else {
		syslog(LOG_NOTICE,
		    "requested file \"%s\" not found: hd unspecified",
		    bp->bp_file);
		return;
	    }
	}
    } else {
	/*
	 * No file specified by the client.
	 */
	if (hp->flags.bootfile && ((hp->bootfile->string)[0] == '/')) {
	    strcpy(path, hp->bootfile->string);
	} else if (hp->flags.homedir && hp->flags.bootfile) {
	    strcpy(path, hp->homedir->string);
	    strcat(path, "/");
	    strcat(path, hp->bootfile->string);
	} else {
	    bzero(bp->bp_file, sizeof(bp->bp_file));
	    goto skip_file;	/* Don't bother trying to access the file */
	}
    }

    /*
     * First try to find the file with a ".host" suffix
     */
    n = strlen(path);
    strcat(path, ".");
    strcat(path, hp->hostname->string);
    if (chk_access(path, &bootsize) < 0) {
	/* If client specified a file, try with "/host" suffix */
	if (bp->bp_file[0]) {
	    path[n] = '/';
	    if (chk_access(path, &bootsize) >= 0) {
		/* XXX too hard to restructure blasted code */
		strcpy(bp->bp_file, path);
		bootsize = 0;
		goto skip_file;
	    }
	}
	path[n] = '\0';			/* Try it without the suffix */
	if (chk_access(path, &bootsize) < 0) {
	    if (bp->bp_file[0]) {
		/*
		 * Client wanted specific file
		 * and we didn't have it.
		 */
		syslog(LOG_NOTICE, "requested file not found: \"%s\"", path);
		return;
	    } else {
		/*
		 * Client didn't ask for a specific file and we couldn't
		 * access the default file, so just zero-out the bootfile
		 * field in the packet and continue processing the reply.
		 */
		bzero(bp->bp_file, sizeof(bp->bp_file));
		goto skip_file;
	    }
	}
    }
    strcpy(bp->bp_file, path);

skip_file:  ;


    /*
     * Zero the vendor information area, except for the magic cookie.
     */
    bzero((bp->bp_vend) + 4, (sizeof(bp->bp_vend)) - 4);

    if (debug > 1) {
	syslog(LOG_DEBUG, "vendor magic field is %d.%d.%d.%d",
	    (int) ((bp->bp_vend)[0]),
	    (int) ((bp->bp_vend)[1]),
	    (int) ((bp->bp_vend)[2]),
	    (int) ((bp->bp_vend)[3]));
    }

    /*
     * If this host isn't set for automatic vendor info then copy the
     * specific cookie into the bootp packet, thus forcing a certain
     * reply format.
     */
    if (!hp->flags.vm_auto) {
	bcopy(hp->vm_cookie, bp->bp_vend, 4);
    }

    /*
     * Figure out the format for the vendor-specific info.
     */
    if (bcmp(bp->bp_vend, vm_rfc1048, 4)) {
	/* Not an RFC1048 bootp client */
	dovend_cmu(bp, hp);
    } else {
	/* RFC1048 conformant bootp client */
	dovend_rfc1048(bp, hp, bootsize);
    }
    sendreply(hp);
}



/*
 * This call checks existance of a file. It returns 0 if the file exists
 * and -1 otherwise.
 */

int
chk_access(path, filesize)
    char *path;
    long *filesize;
{
    struct stat sbuf;

    if (stat(path, &sbuf) < 0)
	    return (-1);
    *filesize = (long) sbuf.st_size;
    return (0);
}



/*
 * Process BOOTREPLY packet (something is using us as a gateway).
 */

void
reply()
{
	if (debug)
	    syslog(LOG_DEBUG, "processing boot reply");
	sendreply(NULL);
}


/*
 * Return our address, hopefully on the same net as addr.
 * (Returns the first random interface address if addr is zero.)
 */
u_long
getmyaddr(saddr)
	u_long saddr;
{
	int maxmatch = 0;
	int len, m;
	register struct ifreq *ifrq, *ifrmax;
	u_long maddr;

	ifrmax = ifrq = &ifreq[0];
	len = ifconf.ifc_len;
	for (; len > 0; len -= sizeof(ifreq[0]), ifrq++) {
		m = nmatch(saddr, ((struct sockaddr_in *)
		    (&ifrq->ifr_addr))->sin_addr.s_addr);
		if (m > maxmatch) {
			maxmatch = m;
			ifrmax = ifrq;
		}
	}
	maddr = ((struct sockaddr_in *)(&ifrmax->ifr_addr))->sin_addr.s_addr;
	if (debug > 1) {
		char xbuf[32];

		strcpy(xbuf, intoa(saddr));
		syslog(LOG_DEBUG, "getmyaddr(%s) chooses %s",
		    xbuf, intoa(maddr));
	}
	return(maddr);
}


/*
 * Send a reply packet to the client. If hp is non-NULL, we are the
 * originator of this reply packe, else we are forwarding it.
 */
void
sendreply(hp)
	struct host *hp;
{
	register struct bootp *bp;
	struct in_addr dst;
	struct sockaddr_in to;
	u_long myaddr;
	char sbuf[128];

	bp = (struct bootp *) buf;
	to = sin;
	to.sin_port = htons(bootpc_port);
	bp->bp_op = BOOTREPLY;
	if (bp->bp_ciaddr) {
		/* Send packet to address the client specified */
		dst.s_addr = bp->bp_ciaddr;
	} else if (bp->bp_giaddr && hp) {
		/* Send packet we originated to gateway */
		dst.s_addr = bp->bp_giaddr;
		to.sin_port = htons(bootps_port);
	} else if (bp->bp_yiaddr) {
		/* Send packet to client on (hopefully) the local wire */
		dst.s_addr = bp->bp_yiaddr;
		if (!setarp(dst.s_addr, bp->bp_chaddr, bp->bp_hlen))
			return;
	} else {
		/* No destination */
		if (hp)
			syslog(LOG_ERR, "bogus entry for %s",
			    hp->hostname->string);
		else
			syslog(LOG_ERR,
			    "bogus request from %s (no destination)",
			    intoa(from.sin_addr.s_addr));
		return;
	}

	/* XXX If dst is us or broadcast, we should probably drop the packet */

	/*
	 * If we're forwarding the reply or there is no gateway,
	 * install our "best" address.
	 */
	myaddr = getmyaddr(bp->bp_yiaddr);
	if (hp == NULL || bp->bp_giaddr == 0)
		bp->bp_giaddr = myaddr;

	/*
	 * If we're originating the reply, we're the server.
	 */
	if (hp)
		bp->bp_siaddr = myaddr;

	to.sin_addr = dst;
	if (sendto(server_s, bp, sizeof(*bp), 0, &to, sizeof(to)) < 0) {
		syslog(LOG_ERR, "sendto: %m");
		return;
	}

	(void)strcpy(sbuf, intoa(bp->bp_yiaddr));
	if (*bp->bp_file) {
		(void)strcat(sbuf, " ");
		(void)strcat(sbuf, bp->bp_file);
	}
	if (hp) {
		(void)strcat(sbuf, " (");
		(void)strcat(sbuf, hp->hostname->string);
		(void)strcat(sbuf, ")");
	} else {
		(void)strcat(sbuf, " (from ");
		(void)strcat(sbuf, intoa(from.sin_addr.s_addr));
		(void)strcat(sbuf, ")");
	}
	syslog(LOG_INFO, "%s", sbuf);
}


/*
 * Forward a client's request to another bootp server.
 * For now, we choose to only forward to one address.
 */
void
forwardrequest(hp)
	register struct host *hp;
{
	register struct bootp *bp;
	struct sockaddr_in to;
	struct in_addr dst;

	bp = (struct bootp *) buf;
	/*
	 * If we're the first hop, install our address in the gateway
	 * slot. If we have more than one address, try to use an
	 * address on the same net as the server we're forwarding to.
	 * This increases the chances that the packet will make it back
	 * to us.
	 */
	/*
	 * XXX Don't install gateway address if there's already one;
	 * At least one version of cisco roms don't increment the hop
	 * counter...
	 */
	if (bp->bp_hops == 0 && bp->bp_giaddr == 0)
		bp->bp_giaddr = getmyaddr(hp->forward->addr[0]);

	/* Watch for loops and other problems. */
	if (++bp->bp_hops > MAXHOPS) {
		syslog(LOG_ERR, "%s hop count exceeded (%d > %d)",
		    hp->hostname->string, bp->bp_hops, MAXHOPS);
		return;
	}

	to = sin;
	to.sin_port = htons(bootps_port);
	dst.s_addr = hp->forward->addr[0];
	to.sin_addr = dst;

	if (sendto(client_s, bp, sizeof(*bp), 0, &to, sizeof(to)) < 0) {
		syslog(LOG_ERR, "forwardreply(): sendto: %m");
		return;
	}
	syslog(LOG_INFO, "%s forwarded to %s",
	    hp->hostname->string, intoa(dst.s_addr));
}



/*
 * Return the number of leading bits matching in the
 * internet addresses supplied.
 */
int
nmatch(a, b)
	register u_long a, b;
{
	register n,m;

	n = 0;
	m = 0;
	do {
		++n;
		m |= 1 << (32 - n);
	} while ((a & m) == (b & m) && n < 32);
	return (n - 1);
}

/*
 * Install an arp cache entry so that IP address 'ia' will be
 * temporarily bound to hardware address 'ha' of length 'len'.
 * Returns true if successful, else false.
 */
int
setarp(addr, ha, len)
	u_long addr;
	byte *ha;
	int len;
{
	struct sockaddr_in *si;

	bzero((caddr_t)&arpreq, sizeof(arpreq));

	arpreq.arp_pa.sa_family = AF_INET;
	si = (struct sockaddr_in *) &arpreq.arp_pa;
	si->sin_addr.s_addr = addr;

	arpreq.arp_flags = ATF_INUSE | ATF_COM;

	bcopy(ha, arpreq.arp_ha.sa_data, len);

	if (ioctl(server_s, SIOCSARP, (caddr_t)&arpreq) < 0) {
		syslog(LOG_ERR, "%s: SIOCSARP: %m%s", intoa(addr),
		    errno == ENETUNREACH ? " (client on wrong ethernet?)" : "");
		return (0);
	}
	return (1);
}



#ifdef DEBUG
/*
 * Dump the internal memory database to bootpd_dump.
 */

void
dumptab()
{
    register int n;
    register struct host *hp;
    register FILE *fp;
    long t;

    /*
     * Open bootpd.dump file.
     */
    if ((fp = fopen(bootpd_dump, "w")) == NULL) {
	syslog(LOG_ERR, "error opening \"%s\": %m", bootpd_dump);
	exit(1);
    }

    t = time(NULL);
    fprintf(fp, "\n# %s\n", Version);
    fprintf(fp, "# %s: dump of bootp server database.\n", bootpd_dump);
    fprintf(fp, "#\n# Dump taken %s", ctime(&t));
    fprintf(fp, "#\n#\n# Legend:\n");
    fprintf(fp, "#\thd -- home directory\n");
    fprintf(fp, "#\tbf -- bootfile\n");
    fprintf(fp, "#\tbs -- bootfile size in 512-octet blocks\n");
    fprintf(fp, "#\tcs -- cookie servers\n");
    fprintf(fp, "#\tds -- domain name servers\n");
    fprintf(fp, "#\tgw -- gateways\n");
    fprintf(fp, "#\tha -- hardware address\n");
    fprintf(fp, "#\thd -- home directory for bootfiles\n");
    fprintf(fp, "#\tht -- hardware type\n");
    fprintf(fp, "#\tim -- impress servers\n");
    fprintf(fp, "#\tip -- host IP address\n");
    fprintf(fp, "#\tlg -- log servers\n");
    fprintf(fp, "#\tlp -- LPR servers\n");
    fprintf(fp, "#\tns -- IEN-116 name servers\n");
    fprintf(fp, "#\trl -- resource location protocol servers\n");
    fprintf(fp, "#\tsm -- subnet mask\n");
    fprintf(fp, "#\tto -- time offset (seconds)\n");
    fprintf(fp, "#\tts -- time servers\n\n\n");

    n = 0;
    for (hp = (struct host *) hash_FirstEntry(nmhashtable); hp != NULL;
	 hp = (struct host *) hash_NextEntry(nmhashtable)) {
	    dump_host(fp, hp);
	    fprintf(fp, "\n");
	    n++;
    }
    fclose(fp);

    syslog(LOG_INFO, "dumped %d entries to \"%s\"", n, bootpd_dump);
}



/*
 * Dump all the available information on the host pointed to by "hp".
 * The output is sent to the file pointed to by "fp".
 */

void
dump_host(fp, hp)
    FILE *fp;
    struct host *hp;
{
    register int i;
    register byte *dataptr;

    if (hp) {
	if (hp->hostname) {
	    fprintf(fp, "%s:", hp->hostname->string);
	}
	if (hp->flags.iaddr) {
	    fprintf(fp, "ip=%s:", intoa(hp->iaddr));
	}
	if (hp->flags.htype) {
	    fprintf(fp, "ht=%u:", (unsigned) hp->htype);
	    if (hp->flags.haddr) {
		fprintf(fp, "ha=%s:", haddrtoa(hp->haddr, hp->htype));
	    }
	}
	if (hp->flags.subnet_mask) {
	    fprintf(fp, "sm=%s:", intoa(hp->subnet_mask));
	}
	if (hp->flags.cookie_server) {
	    fprintf(fp, "cs=");
	    list_ipaddresses(fp, hp->cookie_server);
	    fprintf(fp, ":");
	}
	if (hp->flags.domain_server) {
	    fprintf(fp, "ds=");
	    list_ipaddresses(fp, hp->domain_server);
	    fprintf(fp, ":");
	}
	if (hp->flags.gateway) {
	    fprintf(fp, "gw=");
	    list_ipaddresses(fp, hp->gateway);
	    fprintf(fp, ":");
	}
	if (hp->flags.impress_server) {
	    fprintf(fp, "im=");
	    list_ipaddresses(fp, hp->impress_server);
	    fprintf(fp, ":");
	}
	if (hp->flags.log_server) {
	    fprintf(fp, "lg=");
	    list_ipaddresses(fp, hp->log_server);
	    fprintf(fp, ":");
	}
	if (hp->flags.lpr_server) {
	    fprintf(fp, "lp=");
	    list_ipaddresses(fp, hp->lpr_server);
	    fprintf(fp, ":");
	}
	if (hp->flags.name_server) {
	    fprintf(fp, "ns=");
	    list_ipaddresses(fp, hp->name_server);
	    fprintf(fp, ":");
	}
	if (hp->flags.rlp_server) {
	    fprintf(fp, "rl=");
	    list_ipaddresses(fp, hp->rlp_server);
	    fprintf(fp, ":");
	}
	if (hp->flags.time_server) {
	    fprintf(fp, "ts=");
	    list_ipaddresses(fp, hp->time_server);
	    fprintf(fp, ":");
	}
	if (hp->flags.time_offset) {
	    if (hp->flags.timeoff_auto) {
		fprintf(fp, "to=auto:");
	    } else {
		fprintf(fp, "to=%ld:", hp->time_offset);
	    }
	}
	if (hp->flags.homedir) {
	    fprintf(fp, "hd=%s:", hp->homedir->string);
	}
	if (hp->flags.bootfile) {
	    fprintf(fp, "bf=%s:", hp->bootfile->string);
	}
	if (hp->flags.bootsize) {
	    fprintf(fp, "bs=");
	    if (hp->flags.bootsize_auto) {
		fprintf(fp, "auto:");
	    } else {
		fprintf(fp, "%d:", hp->bootsize);
	    }
	}
	if (hp->flags.name_switch && hp->flags.send_name) {
	    fprintf(fp, "hn:");
	}
	if (hp->flags.vendor_magic) {
	    fprintf(fp, "vm=");
	    if (hp->flags.vm_auto) {
		fprintf(fp, "auto:");
	    } else if (!bcmp(hp->vm_cookie, vm_cmu, 4)) {
		fprintf(fp, "cmu:");
	    } else if (!bcmp(hp->vm_cookie, vm_rfc1048, 4)) {
		fprintf(fp, "rfc1048");
	    } else {
		fprintf(fp, "%d.%d.%d.%d:",
			    (int) ((hp->vm_cookie)[0]),
			    (int) ((hp->vm_cookie)[1]),
			    (int) ((hp->vm_cookie)[2]),
			    (int) ((hp->vm_cookie)[3]));
	    }
	}
	if (hp->flags.generic) {
	    fprintf(fp, "generic=");
	    dataptr = hp->generic->data;
	    for (i = hp->generic->length; i > 0; i--) {
		fprintf(fp, "%02X", (int) *dataptr++);
	    }
	    fprintf(fp, ":");
	}
    }
}



/*
 * Dump an entire struct addr_list of IP addresses to the indicated file.
 *
 * The addresses are printed in standard ASCII "dot" notation and separated
 * from one another by a single space.  A single leading space is also
 * printed before the first adddress.
 *
 * Null lists produce no output (and no error).
 */

void
list_ipaddresses(fp, ipptr)
    FILE *fp;
    struct addr_list *ipptr;
{
    register u_int count;
    register u_long *addrptr;

    if (ipptr) {
	count = ipptr->addrcount;
	addrptr = ipptr->addr;
	if (count-- > 0) {
	    fprintf(fp, "%s", intoa(*addrptr++));
	    while (count-- > 0) {
		fprintf(fp, " %s", intoa(*addrptr++));
	    }
	}
    }
}
#endif /* DEBUG */



/*
 * Insert the CMU "vendor" data for the host pointed to by "hp" into the
 * bootp packet pointed to by "bp".
 */

void
dovend_cmu(bp, hp)
    register struct bootp *bp;
    register struct host *hp;
{
    struct cmu_vend *vendp;
    register struct addr_list *taddr;

    /* Fill in vendor information. Subnet mask, default gateway,
	domain name server, ien name server, time server */
    vendp = (struct cmu_vend *) bp->bp_vend;
    if (hp->flags.subnet_mask) {
	vendp->v_smask = hp->subnet_mask;
	vendp->v_flags |= VF_SMASK;
	if (hp->flags.gateway) {
	    vendp->v_dgate = *hp->gateway->addr;
	}
    }
    if (hp->flags.domain_server) {
	taddr = hp->domain_server;
	if (taddr->addrcount > 0) {
	    vendp->v_dns1 = taddr->addr[0];
	    if (taddr->addrcount > 1) {
		vendp->v_dns2 = taddr->addr[1];
	    }
	}
    }
    if (hp->flags.name_server) {
	taddr = hp->name_server;
	if (taddr->addrcount > 0) {
	    vendp->v_ins1 = taddr->addr[0];
	    if (taddr->addrcount > 1) {
		vendp->v_ins2 = taddr->addr[1];
	    }
	}
    }
    if (hp->flags.time_server) {
	taddr = hp->time_server;
	if (taddr->addrcount > 0) {
	    vendp->v_ts1 = taddr->addr[0];
	    if (taddr->addrcount > 1) {
		vendp->v_ts2 = taddr->addr[1];
	    }
	}
    }
    strcpy(vendp->v_magic, vm_cmu);

    if (debug > 1)
	syslog(LOG_INFO, "sending CMU-style reply");
}



/*
 * Insert the RFC1048 vendor data for the host pointed to by "hp" into the
 * bootp packet pointed by "bp".
 */

void
dovend_rfc1048(bp, hp, bootsize)
    register struct bootp *bp;
    register struct host *hp;
    long bootsize;
{
    int bytesleft, len;
    byte *vp;
    char *tmpstr;

    vp = bp->bp_vend;
    bytesleft = sizeof(bp->bp_vend);	/* Initial vendor area size */
    bcopy(vm_rfc1048, vp, 4);		/* Copy in the magic cookie */
    vp += 4;
    bytesleft -= 4;

    if (hp->flags.time_offset) {
	*vp++ = TAG_TIME_OFFSET;			/* -1 byte  */
	*vp++ = 4;					/* -1 byte  */
	if (hp->flags.timeoff_auto) {
	    insert_u_long(htonl(secondswest), &vp);
	} else {
	    insert_u_long(htonl(hp->time_offset), &vp);		/* -4 bytes */
	}
	bytesleft -= 6;
    }
    if (hp->flags.subnet_mask) {
	*vp++ = TAG_SUBNET_MASK;			/* -1 byte  */
	*vp++ = 4;					/* -1 byte  */
	insert_u_long(hp->subnet_mask, &vp);		/* -4 bytes */
	bytesleft -= 6;					/* Fix real count */
	if (hp->flags.gateway) {
	    insert_ip(TAG_GATEWAY, hp->gateway, &vp, &bytesleft);
	}
    }

    if (hp->flags.bootsize) {
	bootsize = (hp->flags.bootsize_auto) ?
		   (bootsize / 512 + 1) : (hp->bootsize);   /* Round up */
	*vp++ = TAG_BOOTSIZE;
	*vp++ = 2;
	*vp++ = (byte) ((bootsize >> 8) & 0xFF);
	*vp++ = (byte) (bootsize & 0xFF);
	bytesleft -= 4;		/* Tag, length, and 16 bit blocksize */
    }

    if (hp->flags.domain_server) {
	insert_ip(TAG_DOMAIN_SERVER, hp->domain_server, &vp, &bytesleft);
    }
    if (hp->flags.name_server) {
	insert_ip(TAG_NAME_SERVER, hp->name_server, &vp, &bytesleft);
    }
    if (hp->flags.rlp_server) {
	insert_ip(TAG_RLP_SERVER, hp->rlp_server, &vp, &bytesleft);
    }
    if (hp->flags.time_server) {
	insert_ip(TAG_TIME_SERVER, hp->time_server, &vp, &bytesleft);
    }
    if (hp->flags.log_server) {
	insert_ip(TAG_LOG_SERVER, hp->log_server, &vp, &bytesleft);
    }
    if (hp->flags.lpr_server) {
	insert_ip(TAG_LPR_SERVER, hp->lpr_server, &vp, &bytesleft);
    }
    if (hp->flags.cookie_server) {
	insert_ip(TAG_COOKIE_SERVER, hp->cookie_server, &vp, &bytesleft);
    }

    if (hp->flags.name_switch && hp->flags.send_name) {
	/*
	 * Check for room for hostname.  Add 2 to account for
	 * TAG_HOSTNAME and length.
	 */
	len = strlen(hp->hostname->string);
	if ((len + 2) > bytesleft) {
	    /*
	     * Not enough room for full (domain-qualified) hostname, try
	     * stripping it down to just the first field (host).
	     */
	    tmpstr = hp->hostname->string;
	    len = 0;
	    while (*tmpstr && (*tmpstr != '.')) {
		tmpstr++;
		len++;
	    }
	}
	if ((len + 2) <= bytesleft) {
	    *vp++ = TAG_HOSTNAME;
	    *vp++ = (byte) (len & 0xFF);
	    bcopy(hp->hostname->string, vp, len);
	    vp += len;
	}
    }
    if (hp->flags.generic) {
	insert_generic(hp->generic, &vp, &bytesleft);
    }
    if (bytesleft >= 1) {
	*vp = TAG_END;
    }

    if (debug > 1) {
	syslog(LOG_INFO, "sending RFC1048-style reply");
    }
}



/*
 * Compare function to determine whether two hardware addresses are
 * equivalent.  Returns TRUE if "host1" and "host2" are equivalent, FALSE
 * otherwise.
 *
 * This function is used when retrieving elements from the hardware address
 * hash table.
 */

boolean
hwlookcmp(host1, host2)
    struct host *host1, *host2;
{
    if (host1->htype != host2->htype) {
	return FALSE;
    }
    if (bcmp(host1->haddr, host2->haddr, haddrlength(host1->htype))) {
	return FALSE;
    }
    return TRUE;
}




/*
 * Compare function for doing IP address hash table lookup.
 */

boolean
iplookcmp(host1, host2)
    struct host *host1, *host2;
{
    return (host1->iaddr == host2->iaddr);
}



/*
 * Insert a tag value, a length value, and a list of IP addresses into the
 * memory buffer indirectly pointed to by "dest", without going past "stop".
 * "ipst_prt" is a pointer to an ipaddr_t (i.e. a structure containing a
 * pointer to a linked list of IP addresses and a linkcount specifying how
 * many hosts are using this list).
 *
 * This is used to fill the vendor-specific area of a bootp packet in
 * conformance to RFC1048.
 */

void
insert_ip(tag, iplist, dest, bytesleft)
    int tag;
    struct addr_list *iplist;
    byte **dest;
    int *bytesleft;
{
    register u_long *addrptr;
    register u_int addrcount;
    byte *d;

    if (iplist && (*bytesleft >= 6)) {
	d = *dest;				/* Save pointer for later */
	**dest = tag;
	(*dest) += 2;
	(*bytesleft) -= 2;			/* Account for tag and length */
	addrptr = iplist->addr;
	addrcount = iplist->addrcount;
	while ((*bytesleft >= 4) && (addrcount > 0)) {
	    insert_u_long(*addrptr, dest);
	    addrptr++;
	    addrcount--;
	    (*bytesleft) -= 4;			/* Four bytes per address */
	}
	d[1] = (byte) ((*dest - d - 2) & 0xFF);
    }
}



/*
 * Insert generic data into a bootp packet.  The data is assumed to already
 * be in RFC1048 format.  It is inserted using a first-fit algorithm which
 * attempts to insert as many tags as possible.  Tags and data which are
 * too large to fit are skipped; any remaining tags are tried until they
 * have all been exhausted.
 */

void
insert_generic(gendata, buff, bytesleft)
    struct shared_bindata *gendata;
    byte **buff;
    int *bytesleft;
{
    byte *srcptr;
    register int length, numbytes;

    if (gendata) {
	srcptr = gendata->data;
	length = gendata->length;
	while ((length > 0) && (*bytesleft > 0)) {
	    switch (*srcptr) {
		case TAG_END:
		    length = 0;		/* Force an exit on next iteration */
		    break;
		case TAG_PAD:
		    *(*buff)++ = *srcptr++;
		    (*bytesleft)--;
		    length--;
		    break;
		default:
		    numbytes = srcptr[1] + 2;
		    if (*bytesleft >= numbytes) {
			bcopy(srcptr, *buff, numbytes);
			(*buff) += numbytes;
			(*bytesleft) -= numbytes;
		    }
		    srcptr += numbytes;
		    length -= numbytes;
		    break;
	    }
	}
    }
}




/*
 * Convert a hardware address to an ASCII string.
 */

char *
haddrtoa(haddr, htype)
    register byte *haddr;
    byte htype;
{
    static char haddrbuf[3 * MAXHADDRLEN + 1];
    register char *bufptr;
    register u_int count;

    bufptr = haddrbuf;
    for (count = haddrlength(htype); count > 0; count--) {
	sprintf(bufptr, "%x:",	(unsigned) (*haddr++ & 0xFF));
	while (*bufptr != '\0')
		++bufptr;
    }
    --bufptr;
    *bufptr = '\0';		/* kill trailing colon */
    return (haddrbuf);
}



/*
 * Insert the u_long "value" into memory starting at the byte
 * pointed to by the byte pointer (*dest).  (*dest) is updated to
 * point to the next available byte.
 *
 * Since it is desirable to internally store network addresses in network
 * byte order (in struct in_addr's), this routine expects longs to be
 * passed in network byte order.
 *
 * However, due to the nature of the main algorithm, the long must be in
 * host byte order, thus necessitating the use of ntohl() first.
 */

void
insert_u_long(value, dest)
    u_long value;
    byte **dest;
{
    register byte *temp;
    register int n;

    value = ntohl(value);	/* Must use host byte order here */
    temp = (*dest += 4);
    for (n = 4; n > 0; n--) {
	*--temp = (byte) (value & 0xFF);
	value >>= 8;
    }
    /* Final result is network byte order */
}
